#include "system.h"
#include "db.h"
#include "dbutil.h"

/*
 * CMDS.C - command interpreter and utilities for database utility program.
 */

extern strcpy();
extern int strlen();
extern int strcmp();

extern d_print();
extern long *d_alloc();
extern d_free();
extern dbInit();
extern dbSetError();
extern dbTerm();
extern int fLookup();
extern word relDB();
extern int fDtoR();
extern word dbOpen();
extern word rOpen();

extern SYMBOL *uLookup();
extern uEnter();
extern uNewTup();
extern udbDisp();
extern uRelDisp();
extern uSymInit();
extern udbCre();
extern udbOpen();
extern udbClose();
extern uRelCre();
extern uRelOpen();
extern uRelClose();
extern uIndex();
extern uUse();
extern uAdd();
extern uDel();
extern uGoto();
extern uNext();
extern uPrev();
extern uMod();
extern uList();
extern uVerbose();
extern uFind();
extern uAbort();
extern uScan();
extern uScans();
extern uTree();
extern uSyms();
extern uTLock();
extern uTUnlock();
extern uRLock();
extern uRUnlock();
extern uIsTLocked();
extern uIsRLocked();
extern uRTupLocks();
extern bool uChkAct();
extern bool uChkTup();
extern uTupDisp();
extern uPutFld();

#ifdef CPM

struct conBuff {
    char cb_mx;
    char cb_nc;
    char cb_buff[79];
} uLineBuff;

#endif
#ifdef XENIX

#define CMDBUFFLEN	250
char uLineBuff[CMDBUFFLEN];

#endif

char *uLinePtr;

/*
 * uGetLine - read an input line from the terminal.
 */

uGetLine()
{

#ifdef CPM
    uLineBuff.cb_mx = 78;
    bdos(10, &uLineBuff);
    d_crlf();
    uLineBuff.cb_buff[uLineBuff.cb_nc] = '\0';
    uLinePtr = &uLineBuff.cb_buff[0];
#else
    uLineBuff[read(0, &uLineBuff[0], CMDBUFFLEN)] = '\0';
    uLinePtr = &uLineBuff[0];
#endif
}

/*
 * uKeyHit - return true if a key was hit.
 */

bool
uKeyHit()
{
    char ch;

#ifdef CPM
    ch = bdos(6, 0xff);
    if (ch == '\0') {
	return(false);
    }
    if (ch == '\023') {
	while (bdos(6, 0xff) != '\021')
	    ;
	return(false);
    }
    return(true);
#else
    blah blah blah
#endif
}

/*
 * uErr1 - first part of an error message.
 */

uErr1()
{

    d_print("*** ");
}

/*
 * uErr2 - last part of an error message.
 */

uErr2()
{

    d_print(". ***");
    d_crlf();
}

/*
 * uErr - issue an error message.
 */

uErr(mess)
    char *mess;
{

    uErr1();
    d_print(mess);
    uErr2();
}

/*
 * uNErr - issue an error message about a name.
 */

uNErr(name, mess)
    char *name, *mess;
{

    uErr1();
    d_print(name);
    d_print(": ");
    d_print(mess);
    uErr2();
}

/*
 * uTail - complain if anything left on command line.
 */

uTail()
{

    if (*uLinePtr != '\0') {
	uErr("Extraneous characters on command line");
    }
}

#ifdef FIELDS

/*
 * uPrtLng - print out a long int value.
 */

uPrtLng(l)
long l;
{
    char buffer[11];
    register char *p;

    if (l >= 0L) {
	l = -l;
    } else {
	d_print("-");
    }
    p = &buffer[10];
    *p = '\0';
    do {
	--p;
	*p = (- (int) (l % 10L)) + '0';
	l /= 10L;
    } while (l != 0L);
    d_print(p);
}

#endif

/*
 * uHandler - a do-nothing error handler.
 */

uHandler(errCode)
    word errCode;
{
/*
    uErr1();
    d_print("Application error #");
    d_pNum(errCode);
    uErr2();
*/
}

/*
 * uSkpBlnk - skip past blanks in the input line.
 */

uSkpBlnk()
{

    while (*uLinePtr == ' ') {
	++uLinePtr;
    }
}

/*
 * uGetWord - get the next word from the input line.
 */

char *
uGetWord()
{
    register char *w;

    uSkpBlnk();
    if (*uLinePtr == '\0') {
	return(nil);
    } else {
	w = uLinePtr;
	while (*uLinePtr != '\0' && *uLinePtr != ' ' && *uLinePtr != ',') {
	    ++uLinePtr;
	}
	if (*uLinePtr != '\0') {
	    *uLinePtr++ = '\0';
	}
	return(w);
    }
}

/*
 * uConvLong - convert a string to a long int (if possible).
 */

bool
uConvLong(w, pNum)
register char *w;
long *pNum;
{
    static long n;
    static bool isNeg;

    isNeg = false;
    if (*w == '-') {
	w++;
	isNeg = true;
    }
    n = 0L;
    while (*w >= '0' && *w <= '9') {
	n = n * 10L - (*w++ - '0');
    }
    *pNum = isNeg ? n : -n;
    return(*w == '\0');
}

/*
 * uConvTID - convert string to tuple identifier; complain if no good.
 */

bool
uConvTID(w, pt)
    char *w;
    long *pt;
{

    if (uConvLong(w, pt)) {
	if (*pt > 0) {
	    return(true);
	}
	uErr("Tuple identifier must be positive");
    } else {
	uErr("Invalid tuple identifier");
    }
    return(false);
}

/*
 * uGetLong - get a long int from the input line.
 */

bool
uGetLong(pNum)
    long *pNum;
{
    char *w;

    w = uGetWord();
    if (w == nil) {
	return(false);
    } else {
	return(uConvLong(w, pNum));
    }
}

/*
 * uGetNum - get an int from the input line.
 */

bool
uGetNum(pNum)
    int *pNum;
{
    long n;
    bool result;

    result = uGetLong(&n);
    *pNum = (int) n;
    return(result);
}

#ifdef FIELDS

/*
 * uGotNum - check for an optional number on the command line. If there
 *	     is one, return true, else return false.
 */

bool
uGotNum()
{

    uSkpBlnk();
    if (*uLinePtr >= '0' && *uLinePtr <= '9') {
	return(true);
    } else {
	return(false);
    }
}

#ifdef GETFIELD

/*
 * uGotChr - return true if characters left on input line.
 */

bool
uGotChr()
{

    uSkpBlnk();
    return(*uLinePtr != '\0');
}

/*
 * uGetStr - get a string from the input line.
 */

char *
uGetStr()
{
    static char *start, *curr;
    static char quote;

    uSkpBlnk();
    if (*uLinePtr != '\'' && *uLinePtr != '\"') {
	uErr("Missing opening quote");
	*uLinePtr = '\0';
	return(nil);
    } else {
	quote = *uLinePtr;
	++uLinePtr;
	start = uLinePtr;
	curr = uLinePtr;
	while (true) {
	    if (*uLinePtr == quote) {
		++uLinePtr;
		if (*uLinePtr != quote) {
		    break;
		}
	    } else if (*uLinePtr == '\0') {
		uErr("Missing closing quote");
		return(nil);
	    }
	    *curr++ = *uLinePtr++;
	}
	*curr = '\0';
	if (*uLinePtr == ',') {
	    ++uLinePtr;
	}
	return(start);
    }
}

#endif

#endif

/*
 * uCopy - copy a string.
 */

char *
uCopy(p)
    register char *p;
{
    static char *q;

    q = (char *)d_alloc(strlen(p) + 1);
    strcpy(q, p);
    return(q);
}

/*
 * uCAP - capitalize a string.
 */

uCAP(p)
    register char *p;
{

    while (*p != '\0') {
	if (*p >= 'a' && *p <= 'z') {
	    *p -= 32;
	}
	p++;
    }
}

/*
 * uFree - free a string.
 */

uFree(p)
    register char *p;
{

    d_free(p, strlen(p) + 1);
}

/*
 * uHelp - display a list of commands.
 */

#ifdef HELP

uHelp()
{
    d_print("Commands are:");d_crlf();d_crlf();
    d_print("QUIT");d_crlf();
    d_print("DBCREATE <dbName>");d_crlf();
    d_print("DBOPEN <dbName>");d_crlf();
    d_print("DBCLOSE <dbName>");d_crlf();
    d_print("RCREATE <dbName> <relName> <xDrives> <dDrives> <fSize>");
	d_crlf();
    d_print("ROPEN <dbName> [d:]<relName>");d_crlf();
    d_print("RCLOSE <relName>");d_crlf();
    d_print("INDEX <relName> <fieldName> {u|n}");d_crlf();
    d_print("USE [<relName>]");d_crlf();
    d_print("DISplay [{<dbName>|<relName>|<fieldName>|<tupleID>}... ]");
	d_crlf();
    d_print("ADD [value ...]");d_crlf();
    d_print("DELete [<tupleID>]");d_crlf();
    d_print("GOTO <tupleID>");d_crlf();
    d_print("NEXT");d_crlf();
    d_print("PREVious");d_crlf();
    d_print("MODify [<tupleID>] [<fieldName> value ...]");d_crlf();
    d_print("LIST [startTupleID [endTupleID]] [fieldName...]");d_crlf();
    d_print("VERBOSE {0|1|2|3}");d_crlf();
    d_print("FIND <fieldName> <value>");d_crlf();
    d_print("SCAN[{ALL|=}] <fieldName> [<value>]");d_crlf();
    d_print("ABORT");d_crlf();
    d_print("SCANS");d_crlf();
    d_print("TREE <fieldName>");d_crlf();
    d_print("SYMbols");d_crlf();
    d_print("TLOCK [<tupleID>] {r|m} [w]");d_crlf();
    d_print("TUNLOCK [<tupleID>]");d_crlf();
    d_print("RLOCK <relName> {r|m} [w]");d_crlf();
    d_print("RUNLOCK <relName>");d_crlf();
    d_print("ISTLOCKED [<tupleID>]");d_crlf();
    d_print("ISRLOCKED <relName>");d_crlf();
    d_print("RTUPLOCKS <relName>");d_crlf();
}

#endif

#ifdef DISPLAY

/*
 * uDisplay - display some values.
 */

uDisplay()
{
    register char *w;
    static SYMBOL *sy;
    static long t;
    static int f;
    static bool needCRLF;

    needCRLF = false;
    w = uGetWord();
    if (w == nil) {
#ifdef FIELDS
	if (uChkAct()) {
	    uTupDisp(ActTup);
	}
#endif
    } else {
	while (w != nil) {
#ifdef FIELDS
	    if (*w >= '0' && *w <= '9') {
		if (needCRLF) {
		    d_crlf();
		}
		needCRLF = false;
		if (uChkAct() && uConvTID(w, &t)) {
		    uTupDisp(t);
		} else {
		    break;
		}
	    } else {
#endif
		sy = uLookup(w);
		if (sy == nil) {
#ifdef FIELDS
		    if (ActRel != 0 &&
			    (f = fLookup(relDB(ActRel), w)) != 0 &&
			    (f = fDtoR(ActRel, f)) != 0) {
			if (uChkTup(ActTup)) {
			    uPutFld(ActTup, f);
			    needCRLF = true;
			} else {
			    needCRLF = false;
			    break;
			}
		    } else {
#endif
			uNErr(w, "not defined");
			needCRLF = false;
			break;
#ifdef FIELDS
		    }
#endif
		} else {
		    if (needCRLF) {
			d_crlf();
		    }
		    needCRLF = false;
		    if (sy->sy_type == SY_DB) {
#ifdef DISPLAYOTHER
			udbDisp(w, sy->sy_value);
#else
			uNErr(w, "open database");
#endif
		    } else if (sy->sy_type == SY_REL) {
#ifdef DISPLAYOTHER
			uRelDisp(w, sy->sy_value);
#else
			uNErr(w, "open relation");
#endif
		    }
		}
#ifdef FIELDS
	    }
#endif
	    w = uGetWord();
	}
	if (needCRLF) {
	    d_crlf();
	}
    }
}

#endif

/*
 * main - main program for database utility - loop, processing commands.
 */

main(argc, argv)
int argc;
char **argv;
{
    register char *w;

    dbInit();
    dbSetError(uHandler);
    scStack(true);
    SymTab = nil;
    Verbose = 0;
    Scans = nil;
    ScFree = nil;
#ifndef OPENS
    uEnter("db", SY_DB, dbOpen("DB"));
    uEnter("rel", SY_REL, rOpen(1, 'A', "REL"));
    ActRel = 1;
    if (tNext(ActRel) <= 0L) {
	ActTup = 0L;
    } else {
	ActTup = 1L;
    }
#else
    ActRel = 0;
    ActTup = 0L;
#endif
    while (true) {
	d_print(": ");
	uGetLine();
	w = uGetWord();
	if (w == nil) {
	    ;
	} else if (strcmp(w, "quit") == 0) {
	    uTail();
	    break;
#ifdef DBCREATE
	} else if (strcmp(w, "dbcreate") == 0) {
	    udbCre();
#endif
#ifdef OPENS
	} else if (strcmp(w, "dbopen") == 0) {
	    udbOpen();
#endif
#ifdef DBCLOSE
	} else if (strcmp(w, "dbclose") == 0) {
	    udbClose();
#endif
#ifdef RCREATE
	} else if (strcmp(w, "rcreate") == 0) {
	    uRelCre();
#endif
#ifdef OPENS
	} else if (strcmp(w, "ropen") == 0) {
	    uRelOpen();
#endif
#ifdef RCLOSE
	} else if (strcmp(w, "rclose") == 0) {
	    uRelClose();
#endif
#ifdef DISPLAY
	} else if (strcmp(w, "dis") == 0) {
	    uDisplay();
#endif
#ifdef INDEX
	} else if (strcmp(w, "index") == 0) {
	    uIndex();
#endif
#ifdef FIELDS
#ifdef OPENS
#ifdef USE
	} else if (strcmp(w, "use") == 0) {
	    uUse();
#endif
#endif
#ifdef ADD
	} else if (strcmp(w, "add") == 0) {
	    uAdd();
#endif
#ifdef DELETE
	} else if (strcmp(w, "del") == 0) {
	    uDel();
#endif
#ifdef GOTO
	} else if (strcmp(w, "goto") == 0) {
	    uGoto();
#endif
#ifdef NEXT
	} else if (strcmp(w, "next") == 0) {
	    uNext();
#endif
#ifdef PREVIOUS
	} else if (strcmp(w, "prev") == 0) {
	    uPrev();
#endif
#ifdef MODIFY
	} else if (strcmp(w, "mod") == 0) {
	    uMod();
#endif
#ifdef LIST
	} else if (strcmp(w, "list") == 0) {
	    uList();
#endif
	} else if (strcmp(w, "verbose") == 0) {
	    uVerbose();
#ifdef FIND
	} else if (strcmp(w, "find") == 0) {
	    uFind();
#endif
#ifdef SCANS
	} else if (strcmp(w, "scan") == 0) {
	    uScan(0);
	} else if (strcmp(w, "scanall") == 0) {
	    uScan(1);
	} else if (strcmp(w, "scan=") == 0) {
	    uScan(2);
	} else if (strcmp(w, "abort") == 0) {
	    uAbort();
#ifdef TREE
	} else if (strcmp(w, "tree") == 0) {
	    uTree();
#endif
#ifdef SCANLIST
	} else if (strcmp(w, "scans") == 0) {
	    uScans();
#endif
#endif
#endif
#ifdef SYMBOLS
	} else if (strcmp(w, "sym") == 0) {
	    uSyms();
#endif
#ifdef LOCKING
	} else if (strcmp(w, "tlock") == 0) {
	    uTLock();
	} else if (strcmp(w, "tunlock") == 0) {
	    uTUnlock();
	} else if (strcmp(w, "rlock") == 0) {
	    uRLock();
	} else if (strcmp(w, "runlock") == 0) {
	    uRUnlock();
	} else if (strcmp(w, "istlocked") == 0) {
	    uIsTLocked();
	} else if (strcmp(w, "isrlocked") == 0) {
	    uIsRLocked();
	} else if (strcmp(w, "rtuplocks") == 0) {
	    uRTupLocks();
#endif
#ifdef HELP
	} else if (strcmp(w, "?") == 0 || strcmp(w, "help") == 0) {
	    uHelp();
#endif
	} else {
	    uErr("Unknown command");
	}
    }
    dbTerm();
}
