#include "system.h"
#include "db.h"
#include "dbutil.h"

/*
 * UFIELD.C - routines in DBUTIL dealing with individual fields.
 */

int Verbose;

extern int strlen();
extern int strcmp();

extern d_print();
extern int rFldCnt();
extern int chkTuple();
extern tFree();
extern int fLookup();
extern int fDtoR();
extern rUpdate();
extern char *fData();
extern bool gfBool();
extern bool gfShort();
extern bool gfLong();
extern bool gfFloat();
extern bool gfChar();
extern bool gfString();
extern pfBool();
extern pfShort();
extern pfLong();
extern pfFloat();
extern pfChar();
extern pfString();

extern uGetLine();
extern char *uGetWord();
extern bool uConvTID();
extern bool uChkAct1();
extern bool uChkAct();
extern bool uGetNum();
extern bool uGotNum();
extern uErr();
extern uGetTup();
extern uPrtLng();
extern char *uGetStr();
extern bool uGotChr();
extern bool uGetLong();
extern uCAP();
extern bool uScNext();
extern bool uScPrev();
extern uTail();
extern bool uKeyHit();
extern SCAN **uFndScn();
extern bool uSCNP();

/* buffer for large fields: */

#ifdef FIELDS

char uFldBuf[MAXFIELD + 1];
long uFldLong;
short uFldShort;
bool uFldBool;
double uFldFloat;
char *uFldChar;
byte *uFldPtr;

/*
 * uPTNum - print a tuple number header.
 */

uPTNum(t)
long t;
{

    d_print("Tuple ");
    uPrtLng(t);
    d_print(": ");
}

/*
 * uSetFld - setup for receiving a field value.
 */

uSetFld(typ)
byte typ;
{

    switch (typ) {
    case FTBOOL:
	uFldPtr = (byte *)&uFldBool;
	break;
    case FTSHORT:
	uFldPtr = (byte *)&uFldShort;
	break;
    case FTLONG:
	uFldPtr = (byte *)&uFldLong;
	break;
    case FTFLOAT:
	uFldPtr = (byte *)&uFldFloat;
	break;
    case FTCHAR:
    case FTSTRING:
    case FTUCSTR:
	uFldPtr = (byte *)&uFldBuf;
	break;
    }
}

/*
 * uPrtFld - print a field value which is in the appropriate buffer.
 */

uPrtFld(typ, len)
byte typ;
word len;
{

    switch (typ) {
    case FTBOOL:
	d_print(uFldBool ? "T" : "F");
	break;
    case FTSHORT:
	uPrtLng((long) uFldShort);
	break;
    case FTLONG:
	uPrtLng(uFldLong);
	break;
    case FTFLOAT:
	uPrtLng((long) uFldFloat);
	d_print(".");
	break;
    case FTCHAR:
	d_print("\'");
	uFldBuf[len] = '\0';
	d_print(&uFldBuf[0]);
	d_print("\'");
	break;
    case FTSTRING:
    case FTUCSTR:
	d_print("\"");
	d_print(&uFldBuf[0]);
	d_print("\"");
	break;
    }
}

/*
 * uPutFld - print the value of the given field of the current tuple.
 */

uPutFld(t, f)
long t;
int f;
{
    static char *name;
    static word len;
    static byte typ;
    static int i;

    name = fData(relDB(ActRel), fRtoD(ActRel, f), &typ, &len);
    if (Verbose == 3) {
	d_print(name);
	d_print(": ");
    }
    if (len > MAXFIELD) {
	d_crlf();
	uErr("Field too long - not displayed");
	d_crlf();
    } else {
	uSetFld(typ);
	(typ == FTBOOL ?
	    gfBool
	: typ == FTSHORT ?
	    gfShort
	: typ == FTLONG ?
	    gfLong
	: typ == FTFLOAT ?
	    gfFloat
	: typ == FTCHAR ?
	    gfChar
	:
	    gfString)(ActRel, t, f, uFldPtr);
	uPrtFld(typ, len);
	d_print(" ");
    }
}

/*
 * uTupStat - display the status of the given tuple and return false, else
 *	      return true so caller can do display.
 */

bool
uTupStat(tuple)
long tuple;
{

    uPTNum(tuple);
    switch (chkTuple(ActRel, tuple)) {
    case 0:
	return(true);
    case 1:
	d_print("is free\r\n");
	break;
    case 2:
	d_print("has never been allocated\r\n");
	break;
    }
    return(false);
}

/*
 * uTupDisp - display the status/contents of the given tuple.
 */

uTupDisp(tuple)
long tuple;
{
    register word fldId;

    if (uTupStat(tuple)) {
	for (fldId = 1; fldId <= rFldCnt(ActRel); fldId += 1) {
	    uPutFld(tuple, fldId);
	}
	d_crlf();
    }
}

/*
 * uGFName - get a field name.
 */

bool
uGFName(fp)
int *fp;
{
    register int f;
    static char *w;

    w = uGetWord();
    if (w == nil) {
	uErr("Missing field name");
    } else {
	f = fLookup(relDB(ActRel), w);
	if (f == 0) {
	    uErr("No such field");
	} else {
	    f = fDtoR(ActRel, f);
	    if (f != 0) {
		*fp = f;
		return(true);
	    }
	    uErr("Field not in this relation");
	}
    }
    return(false);
}

#ifdef LIST

/*
 * uList - list status/contents of some or all tuples in relation.
 */

uList()
{
    register unsigned int fieldCount;
    static long last, tuple;
    static int field[34];		/* this is max for one input line */
    static int f;
#ifdef SCANS
    static unsigned int verboseSave;
    static long actTupSave;
    static bool scanning;
#endif

    if (uChkAct()) {
#ifdef SCANS
	scanning = true;
#endif
	tuple = 1L;
	last = tNext(ActRel) - 1L;
	if (uGotNum()) {
#ifdef SCANS
	    scanning = false;
#endif
	    if (!uGetLong(&tuple) || tuple <= 0L) {
		uErr("Invalid starting tuple identifier");
		return;
	    }
	}
	if (uGotNum() && (!uGetLong(&last) || last <= 0L)) {
	    uErr("Invalid ending tuple identifier");
	    return;
	}
#ifdef SCANS
	if (scanning) {
	    scanning = *uFndScn() != nil;
	}
#endif
	fieldCount = 0;
	while (uGotChr()) {
	    if (!uGFName(&f)) {
		return;
	    }
	    field[fieldCount++] = f;
	}
#ifdef SCANS
	if (scanning) {
	    verboseSave = Verbose;
	    actTupSave = ActTup;
	}
#endif
	while (true) {
#ifdef SCANS
	    if (scanning) {
		Verbose = 0;
		if (!uSCNP(true, true)) {
		    break;
		}
		tuple = ActTup;
		Verbose = verboseSave;
	    } else {
#endif
		if (tuple > last) {
		    break;
		}
#ifdef SCANS
	    }
#endif
	    if (uKeyHit()) {
		break;
	    }
#ifdef SCANS
	    if (tuple > 0L) {
#endif
		if (fieldCount == 0) {
		    uTupDisp(tuple);
		} else if (uTupStat(tuple)) {
		    for (f = 0; f != fieldCount; f += 1) {
			uPutFld(tuple, field[f]);
		    }
		    d_crlf();
		}
#ifdef SCANS
	    }
	    if (!scanning) {
#endif
		tuple += 1L;
#ifdef SCANS
	    }
#endif
	}
#ifdef SCANS
	if (scanning) {
	    Verbose = verboseSave;
	    ActTup = actTupSave;
	}
#endif
    }
}

#endif
#ifdef NEXT

/*
 * uNewTup - moving to a new tuple, do output depending on Verbose.
 */

uNewTup()
{

    if (Verbose == 1) {
	uPTNum(ActTup);
	d_crlf();
    } else if (Verbose >= 2) {
	uTupDisp(ActTup);
    }
}

/*
 * uVerbose - set the verbosity level.
 */

uVerbose()
{
    static int n;

    if (uGetNum(&n) && n >= 0 && n <= 3) {
	uTail();
	Verbose = n;
    } else {
	uErr("Missing or invalid verbosity level");
    }
}

/*
 * uNext - move to the next tuple.
 */

uNext()
{
    static long last;

    uTail();
    if (uChkAct() && uScNext()) {
	last = tNext(ActRel);
	if (ActTup == last - 1L) {
	    uErr("End of relation");
	} else {
	    ActTup += 1L;
	    uNewTup();
	}
    }
}

#endif

#ifdef PREVIOUS

/*
 * uPrev - move to the previous tuple.
 */

uPrev()
{
    static long one;

    uTail();
    if (uChkAct() && uScPrev()) {
	one = 1L;
	if (ActTup == one) {
	    uErr("Start of relation");
	} else {
	    ActTup -= 1L;
	    uNewTup();
	}
    }
}

#endif

#ifdef GOTO

/*
 * uGoto - go to a specific tuple.
 */

uGoto()
{
    register char *w;
    static long t;

    if (uChkAct()) {
	w = uGetWord();
	if (w == nil) {
	    uErr("Missing tuple identifier");
	} else if (uConvTID(w, &t)) {
	    uTail();
	    if (t >= tNext(ActRel)) {
		uErr("No such tuple");
	    } else {
		ActTup = t;
		uNewTup();
	    }
	}
    }
}

#endif

/*
 * uChkTup - verify that the given tuple ID is useable.
 */

bool
uChkTup(t)
long t;
{
    register unsigned int code;

    code = chkTuple(ActRel, t);
    if (code == 0) {
	return(true);
    }
    uErr(code == 1 ? "Tuple has been freed" : "No such tuple");
    return(false);
}

#ifdef GETFIELD

/*
 * uInFld - input a field value.
 */

bool
uInFld(typ, len)
byte typ;
word len;
{
    register char *p;
    static char *s;

    switch (typ) {
    case FTBOOL:
	p = uGetWord();
	uCAP(p);
	if (strcmp(p, "T") == 0 || strcmp(p, "F") == 0) {
	    uFldBool = *p == 'T';
	    return(true);
	}
	uErr("Illegal bool value");
	break;
    case FTSHORT:
	if (uGetLong(&uFldLong) && uFldLong <= 32767L && uFldLong>=-32768L) {
	    uFldShort = (short) uFldLong;
	    return(true);
	}
	uErr("Illegal short value");
	break;
    case FTLONG:
	if (uGetLong(&uFldLong)) {
	    return(true);
	}
	uErr("Illegal long value");
	break;
    case FTFLOAT:
	s = uGetWord();
	p = s;
	while (*p != '\0' && *p != '.') {
	    ++p;
	}
	*p = '\0';
	if (uConvLong(s, &uFldLong)) {
	    uFldFloat = (double) uFldLong;
	    return(true);
	}
	uErr("Illegal float value");
	break;
    case FTCHAR:
	p = uGetStr();
	if (p != nil) {
	    if (strlen(p) == len) {
		uFldChar = p;
		return(true);
	    }
	    uErr("Wrong length for char value");
	}
	break;
    case FTSTRING:
    case FTUCSTR:
	return((uFldChar = uGetStr()) != nil);
	break;
    }
    return(false);
}

/*
 * uGetFld - get a value for the given field of the active tuple of the
 *	     active relation.
 */

uGetFld(t, f)
long t;
register int f;
{
    static char *name;
    static word len;
    static byte typ;

    name = fData(relDB(ActRel), fRtoD(ActRel, f), &typ, &len);
    while (true) {
	while (!uGotChr()) {
	    d_print(name);
	    d_print(": ");
	    uGetLine();
	}
	if (uInFld(typ, len)) {
	    switch (typ) {
	    case FTBOOL:
		pfBool(ActRel, t, f, uFldBool);
		break;
	    case FTSHORT:
		pfShort(ActRel, t, f, uFldShort);
		break;
	    case FTLONG:
		pfLong(ActRel, t, f, uFldLong);
		break;
	    case FTFLOAT:
		pfFloat(ActRel, t, f, uFldFloat);
		break;
	    case FTCHAR:
		pfChar(ActRel, t, f, uFldChar);
		break;
	    case FTSTRING:
	    case FTUCSTR:
		pfString(ActRel, t, f, uFldChar);
		break;
	    }
	    return;
	}
    }
}

#endif

#ifdef ADD

/*
 * uAdd - add a tuple to the current relation.
 */
    
uAdd()
{
    register int fldId;
    static char *valu;

    if (uChkAct1()) {
	ActTup = tAlloc(ActRel);
	uPTNum(ActTup);
	d_crlf();
	for (fldId = 1; fldId <= rFldCnt(ActRel); fldId += 1) {
	    uGetFld(ActTup, fldId);
	}
	rUpdate(ActRel);
    }
}

#endif

/*
 * uOptTID - get an optional tupleID, else use active one.
 */

bool
uOptTID(gotOne, pt)
bool gotOne;
long *pt;
{

    if (uChkAct()) {
	if (gotOne) {
	    if (!uConvTID(uGetWord(), pt)) {
		return(false);
	    }
	} else {
	    *pt = ActTup;
	}
	if (uChkTup(*pt)) {
	    return(true);
	}
    }
    return(false);
}

#ifdef MODIFY

/*
 * uMod - modify one or more fields in a tuple.
 */

uMod()
{
    static long t;
    static int f;

    if (uOptTID(uGotNum(), &t)) {
	while (uGotChr()) {
	    if (uGFName(&f)) {
		uGetFld(t, f);
	    } else {
		break;
	    }
	}
	rUpdate(ActRel);
    }
}

#endif

#ifdef DELETE

/*
 * uDel - delete a tuple from the current relation.
 */

uDel()
{
    static long t;

    if (uOptTID(uGotChr(), &t)) {
	uTail();
	tFree(ActRel, t);
	rUpdate(ActRel);
	d_print("Deleted\r\n");
    }
}

#endif

#endif
