#include "system.h"
#include "db.h"
#include "dbutil.h"

/*
 * UREL.C - relation handling routines for the database utility program.
 */

extern word relDB();
extern int rFldCnt();
extern d_print();
extern d_pNum();
extern int fRtoD();
extern bool isIndexed();
extern bool isUnique();
extern word rCreate();
extern int fLookup();
extern bool rcField();
extern bool rcEnd();
extern word rOpen();
extern rClose();
extern int fDtoR();
extern char *fData();
extern xCreate();

extern uFldDisp();
extern char *uGetWord();
extern uErr();
extern uCAP();
extern word uGetDB();
extern bool uNotDef();
extern bool uGetNum();
extern char *uCopy();
extern uGetLine();
extern uNErr();
extern uEnter();
extern uFree();
extern SYMBOL *uLookup();
extern uDelete();
extern uNewTup();
extern uScnTrm();
extern uTail();
extern bool uGotChr();

word ActRel;
long ActTup;
char *uRName;

#ifdef DISPLAYOTHER

/*
 * uRelDisp - display a relation.
 */

uRelDisp(name, relId)
    char *name;
    word relId;
{
    register int i;
    static word dbId;
    static int fldCnt;

    dbId = relDB(relId);
    fldCnt = rFldCnt(relId);
    d_print("Relation ");
    d_print(name);
    d_print(" (database ");
    d_print(dbName(dbId));
    d_print(") has ");
    d_pNum(fldCnt);
    d_print(" fields; next tuple is ");
    d_pNum((word) tNext(relId));
    d_crlf();
    for (i = 1; i <= fldCnt; i++) {
	uFldDisp(dbId, fRtoD(relId, i));
	if (isIndexed(relId, i)) {
	    d_print(" [indexed, ");
	    if (isUnique(relId, i)) {
		d_print("unique]");
	    } else {
		d_print("non-unique]");
	    }
	}
	d_crlf();
    }
}

#endif

/*
 * uGRName - get a relation name.
 */

bool
uGRName()
{

    uRName = uGetWord();
    if (uRName == nil) {
	uErr("Missing relation name");
	return(false);
    } else {
	return(true);
    }
}

#ifdef OPENS

/*
 * uChkDrv - check a drive specification, complain if wrong.
 */

bool
uChkDrv(d)
char d;
{
    if (d >= 'A' && d <= 'P') {
	return(true);
    }
    uErr("Invalid drive specification");
    return(false);
}

#ifdef RCREATE

/*
 * uDrives - get a drive specification.
 */

char *
uDrives()
{
    register char *p;
    static char *drives;
    static char bad;

    drives = uGetWord();
    if (drives == nil) {
	uErr("Missing drive sequence");
	return(nil);
    } else {
	uCAP(drives);
	bad = false;
	p = drives;
	while (*p != '\0') {
	    if (*p != '0' && !uChkDrv(*p)) {
		bad = true;
	    }
	    p++;
	}
	return(bad ? nil : drives);
    }
}

/*
 * uRelCre - create a relation.
 */

uRelCre()
{
    static char *temp, *xDrives, *dDrives, *name;
    static int fSize;
    static word dbId, relId, dbFldId;

    dbId = uGetDB();
    if (dbId != 0 && uGRName() && uNotDef(uRName) &&
	    (xDrives = uDrives()) != nil &&
	    (dDrives = uDrives()) != nil) {
	if (!uGetNum(&fSize) || fSize <= 0 || fSize > 8192) {
	    uErr("Missing or invalid file size");
	} else {
	    temp = uCopy(uRName);
	    uCAP(temp);
	    uTail();
	    relId = rCreate(dbId, temp, xDrives, dDrives, fSize);
	    uFree(temp);
	    if (relId == 0) {
		uErr("Can't create relation");
	    } else {
		/* input line will be re-used, so get copy of name: */
		temp = uCopy(uRName);
		d_print("Enter field names:\r\n");
		while (true) {
		    d_print("? ");
		    uGetLine();
		    name = uGetWord();
		if (name == nil) break;
		    uTail();
		    dbFldId = fLookup(dbId, name);
		    if (dbFldId == 0) {
			uNErr(name, "not in this database");
		    } else if (!rcField(dbFldId)) {
			uNErr(name, "already in relation");
		    }
		}
		if (!rcEnd()) {
		    uErr("No fields given for relation");
		} else {
		    uEnter(temp, SY_REL, relId);
		}
		uFree(temp);
	    }
	}
    }
}

#endif

/*
 * uRelOpen - open an existing relation.
 */

uRelOpen()
{
    register char *temp;
    static word dbId, relId;
    static char drive;

    dbId = uGetDB();
    if (dbId != 0 && uGRName()) {
	if (*(uRName + 1) == ':') {
	    *(uRName + 1) = '\0';
	    uCAP(uRName);
	    drive = *uRName;
	    uRName += 2;
	} else {
	    drive = '0';
	}
	uTail();
	if ((drive == '0' || uChkDrv(drive)) && uNotDef(uRName)) {
	    temp = uCopy(uRName);
	    uCAP(temp);
	    relId = rOpen(dbId, drive, temp);
	    if (relId == 0) {
		uErr("Can't open relation");
	    } else {
		uEnter(uRName, SY_REL, relId);
	    }
	    uFree(temp);
	}
    }
}

#endif

/*
 * uGetRel - get a relation id through a name.
 */

word
uGetRel()
{
    register SYMBOL *sy;

    if (!uGRName()) {
	return(0);
    } else {
	sy = uLookup(uRName);
	if (sy == nil || sy->sy_type != SY_REL) {
	    uNErr(uRName, "not a relation");
	    return(0);
	} else {
	    return(sy->sy_value);
	}
    }
}

#ifdef RCLOSE

/*
 * uRelClose - close a relation.
 */

uRelClose()
{
    register word relId;

    relId = uGetRel();
    if (relId != 0) {
	uTail();
#ifdef SCANS
	uScnTrm(relId);
#endif
	rClose(relId);
	uDelete(uRName);
	if (relId == ActRel) {
	    ActRel = 0;
	    ActTup = 0L;
	}
    }
}

#endif

#ifdef INDEX

/*
 * uIndex - create an index for a field of the given relation.
 */

uIndex()
{
    register char *name;
    static word relId, relFldId, len;
    static byte typ;

    relId = uGetRel();
    if (relId != 0) {
	if (tNext(relId) > 1l) {
	    uErr("That relation has existing tuples");
	} else {
	    name = uGetWord();
	    if (name == nil) {
		uErr("Missing field name");
	    } else {
		relFldId = fLookup(relDB(relId), name);
		if (relFldId != 0) {
		    fData(relDB(relId), relFldId, &typ, &len);
		    if (len > MAXKEY) {
			uErr("Field too long to index");
			return;
		    }
		    relFldId = fDtoR(relId, relFldId);
		}
		if (relFldId == 0) {
		    uNErr(name, "not in that relation");
		} else if (isIndexed(relId, relFldId)) {
		    uErr("Field is already indexed");
		} else {
		    name = uGetWord();
		    uTail();
		    if (name == nil || (*name != 'u' && *name != 'n')) {
			uErr("Missing or invalid index type");
		    } else {
			if (!xCreate(relId, relFldId, *name == 'u')) {
			    uErr("Can't create index");
			}
		    }
		}
	    }
	}
    }
}

#endif

#ifdef FIELDS

#ifdef OPENS

#ifdef USE

/*
 * uUse - select a relation to be the active relation.
 */

uUse()
{

    if (uGotChr()) {
	ActRel = uGetRel();
	uTail();
	if (ActRel != 0) {
	    if (tNext(ActRel) <= 0L) {
		ActTup = 0L;
	    } else {
		ActTup = 1L;
		uNewTup();
	    }
	}
    } else {
	ActRel = 0;
	ActTup = 0L;
    }
}

#endif

#endif

/*
 * uChkAct1 - verify that there is an active relation.
 */

bool
uChkAct1()
{

    if (ActRel == 0) {
	uErr("No active relation");
	return(false);
    }
    return(true);
}

/*
 * uChkAct - complain if there is no current active relation with tuples.
 */

bool
uChkAct()
{

    if (uChkAct1()) {
	if (ActTup != 0L) {
	    return(true);
	}
	uErr("Relation has no tuples");
    }
    return(false);
}

#endif
