#include "system.h"
#include "db.h"
#include "dbutil.h"

/*
 * USCAN.C - routines in DBUTIL dealing with index scanning.
 */

extern d_print();
extern long *d_alloc();
extern int fRtoD();
extern word relDB();
extern bool isIndexed();
extern bool scanBool();
extern bool scanShort();
extern bool scanLong();
extern bool scanFloat();
extern bool scanChar();
extern bool scanString();
extern bool fAllBool();
extern bool fAllShort();
extern bool fAllLong();
extern bool fAllFloat();
extern bool fAllChar();
extern bool fAllString();
extern bool strtLow();
extern setAll();
extern fAbort();
extern word scNext();

extern bool uChkAct();
extern uErr();
extern bool uGFName();
extern bool uInFld();
extern uNewTup();
extern uTail();
extern bool uGotChr();
extern bool uKeyHit();

SCAN *Scans;
SCAN *ScFree;

#ifdef FIELDS

#ifdef FIND

/*
 * uFind - simple index lookup.
 */

uFind()
{
    static long t;
    static int f;
    static word len;
    static byte typ;

    if (uChkAct() && uGFName(&f)) {
	if (isIndexed(ActRel, f)) {
	    if (uGotChr()) {
		t = 0L;
		fData(relDB(ActRel), fRtoD(ActRel, f), &typ, &len);
		if (uInFld(typ, len)) {
		    switch (typ) {
		    case FTBOOL:
			t = fOneBool(ActRel, f, uFldBool);
			break;
		    case FTSHORT:
			t = fOneShort(ActRel, f, uFldShort);
			break;
		    case FTLONG:
			t = fOneLong(ActRel, f, uFldLong);
			break;
		    case FTFLOAT:
			t = fOneFloat(ActRel, f, uFldFloat);
			break;
		    case FTCHAR:
			t = fOneChar(ActRel, f, uFldChar);
			break;
		    case FTSTRING:
		    case FTUCSTR:
			t = fOneString(ActRel, f, uFldChar);
			break;
		    }
		    if (t <= 0L) {
			uErr("None found");
		    } else {
			ActTup = t;
			uNewTup();
		    }
		}
		uTail();
	    } else {
		uErr("Missing key value");
	    }
	} else {
	    uErr("Field not indexed");
	}
    }
}

#endif

#ifdef SCANS

/*
 * uScAlloc - allocate a SCAN structure.
 */

SCAN *
uScAlloc()
{
    register SCAN *sc;

    if (ScFree != nil) {
	sc = ScFree;
	ScFree = ScFree->sc_next;
    } else {
	sc = (SCAN *)d_alloc(sizeof(SCAN));
    }
    return(sc);
}

/*
 * uScFree - free a SCAN structure.
 */

uScFree(sc)
register SCAN *sc;
{

    sc->sc_next = ScFree;
    ScFree = sc;
}

/*
 * uStart - start a scan of the indicated field/type.
 */

bool
uStart(typ, f, mode)
byte typ;
register int f;
int mode;
{

    switch (typ) {
    case FTBOOL:
	return((mode == 2 ?   fAllBool :   scanBool)
		    (ActRel, ActRel, f,  uFldBool));
    case FTSHORT:
	return((mode == 2 ?  fAllShort :  scanShort)
		    (ActRel, ActRel, f, uFldShort));
    case FTLONG:
	return((mode == 2 ?   fAllLong :   scanLong)
		    (ActRel, ActRel, f,  uFldLong));
    case FTFLOAT:
	return((mode == 2 ?  fAllFloat :  scanFloat)
		    (ActRel, ActRel, f, uFldFloat));
    case FTCHAR:
	return((mode == 2 ?   fAllChar :   scanChar)
		    (ActRel, ActRel, f,  uFldChar));
    case FTSTRING:
    case FTUCSTR:
	return((mode == 2 ? fAllString : scanString)
		    (ActRel,ActRel, f,  uFldChar));
    }
}

/*
 * uScan - start a scan.
 */

uScan(mode)
int mode;
{
    register SCAN *sc;
    static int f;
    static word len;
    static byte typ;

    if (uChkAct() && uGFName(&f)) {
	if (isIndexed(ActRel, f)) {
	    fData(relDB(ActRel), fRtoD(ActRel, f), &typ, &len);
	    if (uGotChr()) {
		if (!uInFld(typ, len)) {
		    return;
		}
		if (!uStart(typ, f, mode)) {
		    uErr(mode == 2 ? "None found" : "Index empty");
		    return;
		}
	    } else {
		if (mode == 2) {
		    uErr("Missing key value");
		    return;
		}
		if (!strtLow(ActRel, ActRel, f)) {
		    uErr("Index empty");
		    return;
		}
	    }
	    if (mode == 1) {
		setAll(ActRel);
	    }
	    sc = uScAlloc();
	    sc->sc_next = Scans;
	    Scans = sc;
	    sc->sc_rel = ActRel;
	    sc->sc_field = f;
	    sc->sc_mode = mode;
	} else {
	    uErr("Field not indexed");
	}
	uTail();
    }
}

/*
 * uFndScn - find the latest scan on the current relation (if any).
 *	     Return a pointer to the pointer to that scan.
 */

SCAN **
uFndScn()
{
    register SCAN **scp;

    scp = &Scans;
    while (*scp != nil && (*scp)->sc_rel != ActRel) {
	scp = &(*scp)->sc_next;
    }
    return(scp);
}

/*
 * uAbort - abort the latest scan on the current relation.
 */

uAbort()
{
    register SCAN **scp;
    static SCAN *sc;

    uTail();
    if (uChkAct()) {
	scp = uFndScn();
	if (*scp == nil) {
	    uErr("No active scan for this relation");
	} else {
	    sc = *scp;
	    *scp = sc->sc_next;
	    fAbort(ActRel);
	    uScFree(sc);
	    uErr("Scan aborted");
	}
    }
}

/*
 * uSCNP - next or prev on a scan.
 */

bool
uSCNP(isNext, isList)
bool isNext, isList;
{
    register SCAN **scp;
    static SCAN *sc;
    static long t;

    scp = uFndScn();
    if (*scp == nil) {
	return(true);
    }
    t = isNext ? fNext(ActRel) : fPrev(ActRel);
    if (t <= 0L) {
	if (t >= 0L) {
	    sc = *scp;
	    *scp = sc->sc_next;
	    uScFree(sc);
	    uErr("Scan terminated");
	    if (isList) {
		return(false);
	    }
	} else {
	    uErr("Multiple tuples");
	    if (isList) {
		ActTup = 0L;
	    }
	}
    } else {
	ActTup = t;
	uNewTup();
    }
    return(isList);
}

/*
 * uScNext - do a next on a scan if there is one.
 */

bool
uScNext()
{

    return(uSCNP(true, false));
}

/*
 * uScPrev - do a prev on a scan if there is one.
 */

bool
uScPrev()
{

    return(uSCNP(false, false));
}

/*
 * uScnTrm - abort all active scans for the given relation.
 */

uScnTrm(relId)
word relId;
{
    register SCAN **scp;
    static SCAN *sc;

    scp = &Scans;
    while (*scp != nil) {
	sc = *scp;
	if (sc->sc_rel == relId) {
	    fAbort(relId);
	    *scp = sc->sc_next;
	    uScFree(sc);
	} else {
	    scp = &sc->sc_next;
	}
    }
}

#ifdef SCANLIST

/*
 * uScans - print list of active scans.
 */

uScans()
{
    register SCAN *sc;
    static word len;
    static byte typ;

    uTail();
    sc = Scans;
    while (sc != nil) {
	d_print("Scan of relation ");
	d_print(rName(sc->sc_rel));
	d_print(", field ");
	d_print(fData(relDB(sc->sc_rel), fRtoD(sc->sc_rel, sc->sc_field),
		&typ, &len));
	if (sc->sc_mode != 0) {
	    d_print(sc->sc_mode == 1 ? " (all)" : " (=)");
	}
	d_crlf();
	sc = sc->sc_next;
    }
}

#endif

#ifdef TREE

/*
 * uTree - display an index tree.
 */

uTree()
{
    register word depth;
    static long t;
    static int f;
    static word len;
    static byte typ;

    if (uChkAct() && uGFName(&f)) {
	if (isIndexed(ActRel, f)) {
	    fData(relDB(ActRel), fRtoD(ActRel, f), &typ, &len);
	    if (len > MAXFIELD) {
		uErr("Field too long to print");
	    }
	    if (!strtLow(ActRel, ActRel, f)) {
		uErr("Index empty");
		return;
	    }
	    uTail();
	    uSetFld(typ);
	    while (true) {
		if (uKeyHit()) {
		    fAbort(ActRel);
		    break;
		}
		depth = scNext(ActRel, uFldPtr, &t);
	    if (depth == 0) break;
		while (--depth != 0) {
		    d_print("|--");
		}
		uPrtFld(typ, len);
		d_crlf();
	    }
	} else {
	    uErr("Field not indexed");
	}
    }
}

#endif

#endif

#endif
